//
//  kmpg
//
//  Copyright (C) 1997 Bjarni Ivarsson
//  email: Bjarni.Ivarsson@swipnet.se
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//  fileinfo.cpp - Mpeg audio file information

#include <qfile.h>
#include <qlabel.h>
#include <iostream.h>

#include "fileinfo.h"

int tabsel_123[2][3][16] = {
   { {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,},
     {0,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,},
     {0,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,} },

   { {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,},
     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,} }
};

long freqs[7] = { 44100, 48000, 32000, 22050, 24000, 16000 , 11025 };
QString str_freqs[7] = { "44kHz", "48kHz", "32kHz", 
			 "22kHz", "24kHz", "16kHz", "11kHz" };
QString modes[4] = { "Stereo", "Joint-Stereo", 
		     "Dual-Channel", "Single-Channel" };
QString layers[4] = { "Unknown" , "I", "II", "III" };
QString mpeg_type[2] = { "1.0" , "2.0" };
QString id3_genres[80] = { "(none)", "Classic Rock", "Country", "Dance", "Disco", "Funk",
			   "Grunge", "Hip-Hop", "Jazz", "Metal", "New Age",
			   "Oldies", "Other", "Pop", "R&B", "Rap",
			   "Reggae", "Rock", "Techno", "Industrial", "Alternative",
			   "Ska", "Death Metal", "Pranks", "Soundtrack", "Euro-Techno",
			   "Ambient", "Trip-Hop", "Vocal", "Jazz+Funk", "Fusion",
			   "Trance", "Classical", "Instrumental", "Acid", "House",
			   "Game", "Sound Clip", "Gospel", "Noise", "AlternRock",
			   "Bass", "Soul", "Punk", "Space", "Meditative",
			   "Instrumental Pop", "Instrumental Rock", "Ethnic", "Gothic", "Darkwave",
			   "Techno-Industrial", "Electronic", "Pop-Folk", "Eurodance", "Dream",
			   "Southern Rock", "Comedy", "Cult", "Gangsta", "Top 40",
			   "Christian Rap", "Pop/Funk", "Jungle", "Native American", "Cabaret",
			   "New Wave", "Psychadelic", "Rave", "Showtunes", "Trailer",
			   "Lo-Fi", "Tribal", "Acid Punk", "Acid Jazz", "Polka",
			   "Retro", "Musical", "Rock & Roll", "Hard Rock"};

Fileinfo::Fileinfo(KURL url)
{
   int read_bytes=0, i;
   unsigned long header;
   unsigned char buffer[4];

   _is_ok = FALSE;
   _url = url;

   // Currently we can only read local files.
   if( url.protocol() == QString("file") )
   {
      QString fname = url.path();
      QFile file( fname );
      
      _name = fname.copy();
      if (_name.findRev('/')!=-1)
	 _name=_name.right(_name.length()-_name.findRev('/')-1);
   
      if (!file.open(IO_ReadOnly))
      {
	 cerr << "Unable to open file: " << fname << endl;
	 return;
      }

      if (file.size()<4)
      {
	 cerr << "File \"" << fname << "\" is too small to have a header." << endl;
	 return;
      }
   
      while (read_bytes<4)
      {
	 if ((read_bytes+=file.readBlock((char*)buffer+read_bytes, 4-read_bytes))==-1)
	 {
	    cerr << "Unable to read header from file: " << fname << endl;
	    return;
	 }
      }
      
      header=((unsigned long) buffer[0] << 24) | 
	 ((unsigned long) buffer[1] << 16) |
	 ((unsigned long) buffer[2] << 8) | 
	 (unsigned long) buffer[3];

      decodeHeader( header );
      readID3( &file );

      file.close();      
   }
}

Fileinfo::~Fileinfo()
{
}

void Fileinfo::decodeHeader(unsigned long header)
{
   if ((header & 0xffe00000) != 0xffe00000)
   {
      cerr << "Illegal MPEG header found." << endl;
      return;
   }
   
   if (header & (1<<20))
   {
      _lsf = (header & (1<<19))?0:1;
      _mpeg25 = 0;
   }
   else
   {
      _lsf = 1;
      _mpeg25 = 1;
   }
   
   _lay=4-((header>>17)&3);
   _bitrate_index=((header>>12)&0xf);

   if (((header>>10)&0x3) == 0x3)
   {
      cerr << "Stream error" << endl;
      return;
   }

   if (_mpeg25)
      _sampling_frequency = 6;
   else
      _sampling_frequency = ((header>>10)&0x3) + _lsf*3;

   _error_protection = ((header>>16)&0x1)^0x1;
   _padding = ((header>>9)&0x1);
   _extension = ((header>>8)&0x1);
   _copyright = ((header>>3)&0x1);
   _original = ((header>>2)&0x1);
   _emphasis = header & 0x3;
   _mode = ((header>>6)&0x3);
   _mode_ext = ((header>>4)&0x3);
   _padding = ((header>>9)&0x1);
   _stereo = (_mode == 3) ? 1 : 2;
   
   switch(_lay)
   {
      case 1:
	 _framesize = (long) tabsel_123[_lsf][0][_bitrate_index] * 12000;
	 _framesize /= freqs[_sampling_frequency];
	 _framesize = ((_framesize+_padding)<<2);
	 break;
      case 2:
	 _framesize = (long) tabsel_123[_lsf][1][_bitrate_index] * 144000;
	 _framesize /= freqs[_sampling_frequency];
	 _framesize += _padding;
	 break;
      case 3:
	 _framesize = (long) tabsel_123[_lsf][2][_bitrate_index] * 144000;
	 _framesize /= freqs[_sampling_frequency]<<(_lsf);
	 _framesize += _padding;
	 break;
   }
   _is_ok = TRUE;
}

void Fileinfo::readID3(QFile *file)
{
   char str[31];
   char id3_data[128];

   _id3_ok = FALSE;
   _filesize = file->size();   
   file->at( file->size()-128 );
   file->readBlock( id3_data, 128 );
   if ( (id3_data[0]=='T') && (id3_data[1]=='A') && (id3_data[2]=='G') )
   {
      _id3_ok = TRUE;
      _id3_genre = id3_data[127];
      id3_data[127] = '\0';
      _id3_comment = QString( id3_data+97 ).copy();
      id3_data[97] = '\0';
      _id3_year = QString( id3_data+93 ).copy();
      id3_data[93] = '\0';
      _id3_album = QString( id3_data+63 ).copy();
      id3_data[63] = '\0';
      _id3_artist = QString( id3_data+33 ).copy();
      id3_data[33] = '\0';
      _id3_songname = QString( id3_data+3 ).copy();
      
      cerr << "Songname :" << _id3_songname << endl;
      cerr << "Artist   :" << _id3_artist << endl;
      cerr << "Album    :" << _id3_album << endl;
      cerr << "Year     :" << _id3_year << endl;
      cerr << "Comment  :" << _id3_comment << endl;
      cerr << "Genre    :" << (int)_id3_genre << " ";
      cerr << id3_genres[(_id3_genre>79)?0:_id3_genre] << endl;

      _id3_ok = TRUE;
   }
}

int Fileinfo::isOk()
{
   return _is_ok;
}

QString Fileinfo::getName()
{
   return _name;
}

float Fileinfo::getMsPerFrame()
{
   //return ((_lay == 1)?384000.0:1152000.0)/freqs[_sampling_frequency];
   return (float)8*_framesize/tabsel_123[_lsf][_lay-1][_bitrate_index];
}

int Fileinfo::getCopyright()
{
   return _copyright;
}

int Fileinfo::getOriginal()
{
   return _original;
}

QString Fileinfo::getVersion()
{
   return mpeg_type[_lsf];
}

QString Fileinfo::getLayer()
{
   return layers[_lay];
}

QString Fileinfo::getFrequency()
{
   return str_freqs[_sampling_frequency].copy();
}

QString Fileinfo::getMode()
{
   return modes[_mode];
}

int Fileinfo::getChannels()
{
   return _stereo;
}

QString Fileinfo::getBitrate()
{
   QString ret="";
   
   ret.setNum(tabsel_123[_lsf][_lay-1][_bitrate_index]);
   ret = ret + "kbps";
   return ret.copy();
}

int Fileinfo::getFramesize()
{
   return _framesize;
}

int Fileinfo::getMaxNumberOfFrames()
{
   return _filesize/(_framesize-_padding);
}

int Fileinfo::getMinNumberOfFrames()
{
   return _filesize/(_framesize+1-_padding);
}

QString Fileinfo::getID3Songname()
{
   if ( isID3Ok() )
      return _id3_songname.copy();
   else
      return QString("");
}

QString Fileinfo::getID3Artist()
{
   if ( isID3Ok() )
      return _id3_artist.copy();
   else
      return QString("");
}

QString Fileinfo::getID3Album()
{
   if ( isID3Ok() )
      return _id3_album.copy();
   else
      return QString("");
}

QString Fileinfo::getID3Year()
{
   if ( isID3Ok() )
      return _id3_year.copy();
   else
      return QString("");
}

QString Fileinfo::getID3Comment()
{
   if ( isID3Ok() )
      return _id3_comment.copy();
   else
      return QString("");
}

bool Fileinfo::isID3Settable()
{
   bool ret = FALSE;

   if ( !isOk() )
      return ret;

   // We can only set the tag on local files
   if( _url.protocol() == QString("file") )
   {
      QFile file( _url.path() );
      
      ret = file.open( IO_ReadWrite ); // Attempt to open file for write
      file.close();
   }
   return ret;
}

bool Fileinfo::setID3Tag( QString songname, QString artist, QString album,
			  QString year, QString comment, unsigned char genre )
{
   char id3_data[128];
   int i;
   QFile file;

   if ( !isID3Settable() )
      return FALSE;

   _id3_songname = songname.copy();
   _id3_artist = artist.copy();
   _id3_album = album.copy();
   _id3_year = year.copy();
   _id3_comment = comment.copy();
   _id3_genre = genre;

   // Initilize the data buffer to all spaces
   for( i=0; i<128; i++ )
      id3_data[i] = ' ';

   memcpy( id3_data, "TAG", 3 ); // The magic tag
   memcpy( id3_data+3, _id3_songname.data(), _id3_songname.length()<31?_id3_songname.length():30 );
   memcpy( id3_data+33, _id3_artist.data(), _id3_artist.length()<31?_id3_artist.length():30 );
   memcpy( id3_data+63, _id3_album.data(), _id3_album.length()<31?_id3_album.length():30 );
   memcpy( id3_data+93, _id3_year.data(), _id3_year.length()<5?_id3_year.length():4 );
   memcpy( id3_data+97, _id3_comment.data(), _id3_comment.length()<31?_id3_comment.length():30 );
   id3_data[127] = _id3_genre;
   
   file.setName( _url.path() );
   file.open( IO_ReadWrite );

   // Position file pointer correctly
   if ( isID3Ok() )
      file.at( file.size() - 128 );
   else
      file.at( file.size() );

   // Write the new ID3 data to the file
   file.writeBlock( id3_data, 128 );
   _filesize = file.size();
   file.close();

   _id3_ok = TRUE;

   return TRUE;
}

FileinfoDlg::FileinfoDlg( Fileinfo *finfo ) :
   QDialog( 0, "Fileinfo", TRUE ) // FIXME: the dialog is modal
{
   QLabel *lbl[12], *lbl2[6];
   QString str[12] = { "Version :", "Mode :", "Bit Rate :",
		       "Layer :", "Freq :", "Time :",
		       "Name :", "Artist :", "Album :", 
		       "Comment :", "Year :", "Genre :" };
   QString str2[6], string;
   int i, width = 0, width2 = 0, height, sec;

   _finfo = finfo;
   
   // Fill in the MPEG header fields
   str2[0] = _finfo->getVersion();
   str2[1] = _finfo->getMode();
   str2[2] = _finfo->getBitrate();
   str2[3] = _finfo->getLayer();
   str2[4] = _finfo->getFrequency();
   sec = (int)(_finfo->getMinNumberOfFrames()*
	       _finfo->getMsPerFrame()/
	       1000+0.5);
   str2[5].setNum( sec/60 );
   str2[5] += ":";
   if( (sec%60)<10 )
      str2[5] += "0";
   string.setNum( sec%60 );
   str2[5] += string;

   grp1 = new QButtonGroup( "MPEG Header", this );
   grp2 = new QButtonGroup( "ID3 Tag", this );

   // create all labels used
   for( i=0; i<12; i++ )
   {
      lbl[i] = new QLabel( str[i], (i<6?grp1:grp2) );
      lbl[i]->adjustSize();
      if( i < 6 )
      {
	 lbl2[i] = new QLabel( str2[i], grp1 );
	 lbl2[i]->adjustSize();
	 width2 = QMAX( lbl2[i]->width(), width2 );
      }
      width = QMAX( lbl[i]->width(), width );
   }

   // move and size all labels
   height = lbl[0]->height() + 4;
   for( i=0; i<12; i++ )
   {
      if( i < 3 )
	 lbl[i]->move( width - lbl[i]->width()+5, 20+i*height );
      else if( i > 2 && i < 6 )
	 lbl[i]->move( 2*width+width2+10-lbl[i]->width(), 20+(i-3)*height );
      else if( i < 11 )
	 lbl[i]->move( width - lbl[i]->width()+5, 20+(i-6)*(height+8) );
      else
	 lbl[i]->move( width*2, 20+(i-7)*(height+8) );
      if( i < 3 )
	 lbl2[i]->move( width + 10, 20+i*height );
      else if( i < 6 )
	 lbl2[i]->move( 2*width+width2+15, 20+(i-3)*height );
   }

   // create and place the lineedit widgets
   for( i=0; i<5; i++ )
   {
      le[i] = new QLineEdit( grp2 );
      le[i]->setMaxLength( (i<4?30:4) );
      le[i]->adjustSize();
      le[i]->setGeometry( width + 10, 20 + i*(height+8)-4,
			  (i<4?2*width2+width+5:width-15), 
			  le[i]->height() );
      le[i]->setEnabled( _finfo->isID3Settable() );
   }

   // fill in the lineedit widgets
   le[0]->setText( _finfo->getID3Songname() );
   le[1]->setText( _finfo->getID3Artist() );
   le[2]->setText( _finfo->getID3Album() );
   le[3]->setText( _finfo->getID3Comment() );
   le[4]->setText( _finfo->getID3Year() );

   // create, place and fill in the genre combobox
   cbGenre = new QComboBox( FALSE, grp2 );
   cbGenre->adjustSize();
   cbGenre->move( lbl[11]->x()+lbl[11]->width()+5, 20 + 4*(height+8)-4 );
   cbGenre->resize( 2*(width+width2)+15 - cbGenre->x(), le[5]->height() );
   for( i=0; i<80; i++ )
      cbGenre->insertItem( id3_genres[i] );
   cbGenre->setCurrentItem( _finfo->getID3Genre() );

   grp1->setGeometry( 10, 8, 
		      2*(width+width2)+20, 28+3*height );   
   grp2->setGeometry( 10, grp1->y() + grp1->height() + 8,
		      grp1->width(), cbGenre->y() + cbGenre->height() + 5 );

   // add buttons
   pbOk = new QPushButton( "OK", this );
   pbApply = new QPushButton( "Apply", this );
   pbCancel = new QPushButton( "Cancel", this );
   pbOk->adjustSize();
   pbApply->adjustSize();
   pbCancel->adjustSize();
   pbOk->setDefault( TRUE );

   // compute largest button width ( min 75 pixels )
   width = QMAX( pbOk->width(), pbCancel->width() );
   width = QMAX( pbApply->width(), width );
   width = QMAX( width, 75 );

   pbCancel->setGeometry( 10 + grp2->width() - width, 
			  grp2->y() + grp2->height() + 10,
			  width, 
			  pbCancel->height() );
   pbApply->setGeometry( pbCancel->x() - width - 6, pbCancel->y(),
			  width, pbApply->height() );
   pbOk->setGeometry( pbApply->x() - width - 6, pbApply->y(),
			  width, pbOk->height() );

   connect( pbOk, SIGNAL(clicked()), SLOT(okPressed()) );
   connect( pbApply, SIGNAL(clicked()), SLOT(applyPressed()) );
   connect( pbCancel, SIGNAL(clicked()), SLOT(reject()) );

   setFixedSize( grp2->width() + 20, pbOk->y() + pbOk->height() + 8 );
}

FileinfoDlg::~FileinfoDlg()
{
}

void FileinfoDlg::okPressed()
{
   applyPressed();
   accept();
}

void FileinfoDlg::applyPressed()
{
   // ok, now lets set the ID3 tag
   _finfo->setID3Tag( le[0]->text(),            // songname
		      le[1]->text(),            // artist
		      le[2]->text(),            // album
		      le[4]->text(),            // year
		      le[3]->text(),            // comment
		      cbGenre->currentItem() ); // genre
}




